import time
import logging
import requests
import pandas as pd
from datetime import datetime, timezone
from typing import Optional, List


bids_columns = [
     "relay", 
     "timestamp",
     "timestamp_ms",
     "slot", 
     "block_hash", 
     "builder_pubkey", 
     "proposer_pubkey",
     "proposer_fee_recipient",
     "value",
     "gas_used", 
     "gas_limit",
     "block_number",
     "num_tx",
     "optimistic_submission"
    ]

payloads_columns = [
     "relay", 
     "slot", 
     "block_hash", 
     "builder_pubkey", 
     "proposer_pubkey",
     "proposer_fee_recipient",
     "value",
     "gas_used", 
     "gas_limit",
     "block_number",
     "num_tx"
    ]

HEADERS = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36'}

path = "/relay/v1/data/bidtraces/{}?"
fb = "https://boost-relay.flashbots.net" + path
#et = "https://bloxroute.ethical.blxrbdn.com" + path
mp = "https://bloxroute.max-profit.blxrbdn.com" + path
mr = "https://bloxroute.regulated.blxrbdn.com" + path
mf = "https://mainnet-relay.securerpc.com" + path
ed = "https://relay.edennetwork.io" + path
#bn = "https://builder-relay-mainnet.blocknative.com" + path
#rl = "https://relayooor.wtf" + path
ul = "https://relay-analytics.ultrasound.money" + path
ag = "https://agnostic-relay.net" + path
ae = "https://aestus.live" + path
ti = "https://titanrelay.xyz" + path

urls = {
    "bloxroute (max profit)": mp,
    "flashbots": fb,
    #"blocknative": bn,
    "ultra sound": ul,
    "bloxroute (regulated)": mr,
    "aestus": ae,
    "agnostic gnosis": ag,
    #"bloxroute (ethical)": et,
    #"relayooor": rl ,
    "eden": ed,
    "manifold": mf,
    "titan": ti
}

minblocks_relay = {
    "bloxroute (max profit)": 4700936,
    "flashbots": 4700567,
    #"blocknative": 4710139,
    "ultra sound": 5216345,
    "bloxroute (regulated)": 4701043,
    "aestus": 5303933,
    "agnostic gnosis": 5238007,
    #"bloxroute (ethical)": 4702718,
    #"relayooor":5005203 ,
    "eden": 4700737,
    "manifold":4724548,
    "titan": 8280205
}

ALL_RELAYS = ",".join(list(urls.keys()))

class MevBoostCaller:
    def __init__(self, relays: str = ALL_RELAYS) -> None:
        self.relays = relays
        self.endpoints = [RelayEndpoint(relay.strip()) for relay in relays.split(",")]
        
    def get_bids_over_range(self, slots: int, relays: List[str] = ALL_RELAYS, orderby: str = "relay"):
        result = []
        for slot in slots:
            result.append(get_bids(slot, relays, orderby))
        return pd.concat(result, ignore_index=True)
           
    def get_bids(self, slot: int, relays: List[str] = ALL_RELAYS, orderby: str = "relay"):
        bids = []
        for ep in self.endpoints:
            res = ep._get_bids(slot)
            for r in res:
                row = ep._fetch_bid_row(r)
                bids.append(row)
                
        if bids:
            df = pd.DataFrame(bids, columns=bids_columns)
            if orderby:
                df.sort_values(orderby, inplace=True)
            return df.reset_index(drop=True)
        else:
            return pd.DataFrame()
        
    def get_payloads_over_range(self, slots: int, relays: List[str] = ALL_RELAYS, limit: int = 100, orderby: str = "relay"):
        result = []
        for slot in slots:
            result.append(get_payloads(slot, relays, limit, orderby))
        return pd.concat(result, ignore_index=True)
        
    def get_payloads(self, slot: int, relays: List[str] = ALL_RELAYS, limit: int = 100, orderby: str = "relay"):
        payloads = []
        for ep in self.endpoints:
            res = ep._get_payloads(slot, limit = limit)
            for r in res:
                row = ep._fetch_payload_row(r)
                payloads.append(row)
                
        if payloads:
            df = pd.DataFrame(payloads, columns=payloads_columns)
            df = df[df["slot"] == slot]
            if orderby:
                df.sort_values(orderby, inplace=True)
            return df.reset_index(drop=True)
        else:
            return pd.DataFrame()


class RelayEndpoint:
    def __init__(self, name: str) -> None:
        self.name = name
        self.url = urls.get(name)
        self.minslot = minblocks_relay.get(name)


    def _get_bids(self, slot: int, retries: int = 3):
        if self.minslot > slot:
            logging.info(f"Relay not yet active at slot {slot}")
        logging.info(self.url.format("builder_blocks_received") + f"slot={slot}")
        res = requests.get(self.url.format("builder_blocks_received") + f"slot={slot}", timeout=20, headers=HEADERS)
        if not res.status_code == 200:
            if retries == 0:
                return None
            else:
                logging.info('Something for {self.name} failed: Response Status Code != 200; \n'+
                             f"Request: {self.url.format('proposer_payload_delivered')}" + f"cursor={slot}" + limit)
                return None
            return self._get_bids(slot, retries-1)
        return eval(res.content.decode("utf-8").replace("false", "False").replace("true", "True"))
    
    def _get_payloads(self, slot: int, limit: int = None, retries: int = 3):
        if self.minslot > slot:
            logging.info(f"Relay not yet active at slot {slot}")
        if limit:
            limit = f"&limit={limit}"
        else:
            limit = ""
        logging.info(self.url.format("proposer_payload_delivered") + f"cursor={slot}" + limit)
        res = requests.get(self.url.format("proposer_payload_delivered") + f"cursor={slot}" + limit, timeout=20, headers=HEADERS)
        if not res.status_code == 200:
            time.sleep(5)
            if retries == 0:
                return None
            else:
                logging.info('Something for {self.name} failed: Response Status Code != 200; \n'+
                             f"Request: {self.url.format('proposer_payload_delivered')}" + f"cursor={slot}" + limit)
            return self._get_payloads(slot, limit, retries-1)
        return eval(res.content.decode("utf-8").replace("false", "False").replace("true", "True"))
    
    def _fetch_bid_row(self, r, optimistic=False) -> list:
        row = [
            self.name, 
            int(r["timestamp"]),
            int(r["timestamp_ms"]),
            int(r["slot"]),
            r["block_hash"],
            r["builder_pubkey"],
            r["proposer_pubkey"],
            r["proposer_fee_recipient"],
            float(r["value"]),
            int(r["gas_used"]),
            int(r["gas_limit"]),
            int(r["block_number"]),
            int(r["num_tx"])
        ]
        if "optimistic_submission" in r:
            row.append(1 if r["optimistic_submission"] else 0)
        else:
            row.append(0)
        return row    
    
    def _fetch_payload_row(self, r) -> list:
        return [self.name, 
                int(r["slot"]),
                r["block_hash"],
                r["builder_pubkey"],
                r["proposer_pubkey"],
                r["proposer_fee_recipient"],
                float(r["value"]),
                int(r["gas_used"]),
                int(r["gas_limit"]),
                int(r["block_number"]),
                int(r["num_tx"])
               ]


import os
import ast
import json
import time
import logging
import inspect
import textwrap
from io import StringIO
from pathlib import Path
from functools import wraps
from typing import Optional, List, Dict, Any, Callable, TypeVar, Tuple, Union

import requests
from requests.auth import HTTPBasicAuth
import pandas as pd
from tqdm.auto import tqdm

from pyxatu.utils import CONSTANTS
from pyxatu.helpers import PyXatuHelpers
from pyxatu.docscraper import DocsScraper
from pyxatu.client import ClickhouseClient
from pyxatu.retriever import DataRetriever
from pyxatu.validators import ValidatorGadget
from pyxatu.relayendpoint import MevBoostCaller


def column_check_decorator(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # Extract columns and table name from the function arguments
        columns = kwargs.get('columns', "*")
        table = kwargs.get('data_table')

        if table is None:
            raise ValueError("Table name must be provided to verify columns.")
        
        # Perform column verification using the verify_columns method
        if not self.verify_columns(columns, table):
            raise ValueError(f"One or more columns in '{columns}' are not valid for table '{table}'")

        # Call the original function if columns are valid
        return func(self, *args, **kwargs)

    return wrapper

class PyXatu:
    def __init__(
        self, 
        config_path: Optional[str] = None, 
        use_env_variables: bool = False, 
        log_level: str = 'INFO', 
        relay: str = None
    ) -> None:
        if not logging.getLogger().hasHandlers():
            logging.basicConfig(level=getattr(logging, log_level.upper()), format='%(asctime)s - %(levelname)s - %(message)s')
        
        default_path = os.path.join(Path.home(), '.pyxatu_config.json')
        if not config_path is None and not os.path.isfile(default_path):
            raise ValueError("~/.pyxatu_config.json file not found\nRun `xatu setup` to copy the default config file to your HOME directory and then add your credentials to it. Alternatively you can use environment variables.")
        
        if config_path is None:
            self.config_path = default_path
        else:
            self.config_path = config_path
        
        if use_env_variables:
            config = self.read_clickhouse_config_from_env()
            self.clickhouse_url, self.clickhouse_user, self.clickhouse_password = config
        else:
            print("Config Path: ", self.config_path)
            assert os.path.isfile(self.config_path) == True, "Config file not found."
            config = self.read_clickhouse_config_locally()
            self.clickhouse_url, self.clickhouse_user, self.clickhouse_password = config

        logging.info(f"Clickhouse URL: {self.clickhouse_url}, User: {self.clickhouse_user}")
        self.client = ClickhouseClient(self.clickhouse_url, self.clickhouse_user, self.clickhouse_password)
        
        self.data_retriever = DataRetriever(
            client=self.client,
            tables=CONSTANTS["TABLES"]
        )
        
        self.mevboost = MevBoostCaller()
        self.helpers = PyXatuHelpers()
        
        self.method_table_mapping = self.create_method_table_mapping()
        
        self.update_all_column_docs()
        
        
    def read_clickhouse_config_from_env(self) -> Tuple[str, str, str]:
        """Reads Clickhouse configuration from environment variables."""
        clickhouse_user = os.getenv("CLICKHOUSE_USER", "default_user")
        clickhouse_password = os.getenv("CLICKHOUSE_PASSWORD", "default_password")
        url = os.getenv("CLICKHOUSE_URL", "http://localhost")

        logging.info("Clickhouse configs set")
        return url, clickhouse_user, clickhouse_password

    def read_clickhouse_config_locally(self) -> Tuple[str, str, str]:
        """Reads Clickhouse configuration from the config file."""
        with open(self.config_path, 'r') as file:
            config = json.load(file)
        clickhouse_user = config.get("CLICKHOUSE_USER", "default_user")
        clickhouse_password = config.get("CLICKHOUSE_PASSWORD", "default_password")
        url = config.get("CLICKHOUSE_URL", "http://localhost")

        logging.info("Clickhouse configs set")
        return url, clickhouse_user, clickhouse_password
    
    def execute_query(self, query: str, columns: Optional[str] = "*", time_interval: Optional[str] = None) -> Any:
        return self.client.execute_query(query, columns)
    
    @property
    def validators(self):
        if not hasattr(self, '_validators'):
            self._validators = ValidatorGadget()
        return self._validators
    
    @property
    def docs(self):
        if not hasattr(self, '_docs'):
            self._docs = DocsScraper()
        return self._docs
    
    def get_docs(self, table_name: str = None, print_loading: bool = True):
        """
        Retrieves table information, such as available columns, from the DocsScraper.
        """
        if table_name in self.method_table_mapping.keys():
            table_name = self.method_table_mapping[table_name]
        if print_loading:
            print(f"Retrieving schema for table {table_name}")
        return self.docs.get_table_info(table_name)
    
    def _get_types(self, arguments: List[str]) -> List[type]:
        """
        Returns a list of types corresponding to the given argument names based on predefined type hints.

        :param arguments: List of argument names as strings.
        :return: List of types corresponding to the given arguments.
        :raises ValueError: If an argument does not match the predefined list.
        """

        # Define a dictionary mapping argument names to their expected types
        argument_types = {
            "data_table": str,
            "slot": [list, int, type(None)],
            "columns": [str, type(None)],
            "where": [str, type(None)],
            "time_interval": [str, type(None)],
            "network": str,
            "max_retries": int,
            "groupby": [str, type(None)],
            "orderby": [str, type(None)],
            "final_condition": [str, type(None)],
            "limit": [int, type(None)],
            "store_result_in_parquet": [bool, type(None)],
            "custom_data_dir": [str, type(None)],
            "add_final_keyword_to_query": [bool, type(None)]
        }

        types_list = []

        # Iterate over the provided arguments and get the corresponding type
        for arg in arguments:
            if arg in argument_types:
                types_list.append(argument_types[arg])
            else:
                (f"Argument '{arg}' is not recognized.")

        return types_list

    
    @column_check_decorator
    def _get_data(self, *args, **kwargs):
        self.helpers.check_types(kwargs.values(), self._get_types(kwargs.keys()))
        return self.data_retriever.get_data(*args, **kwargs)

    def _generic_getter(self, table_name: str, **kwargs) -> Any:
        return self._get_data(data_table=table_name, **kwargs)

    def get_blockevent(self, **kwargs):
        return self._generic_getter('beacon_api_eth_v1_events_block', **kwargs)
     
    def get_attestation(self, **kwargs) -> Any:
        res = self._generic_getter('canonical_beacon_elaborated_attestation', **kwargs)
        if "validators" in set(res.columns):
            res["validators"] = res["validators"].apply(lambda x: eval(x))
            res = res.explode("validators").reset_index(drop=True)
        return res    
 
    def get_attestation_event(self, add_final_keyword_to_query: bool = False, **kwargs) -> Any:
        kwargs["add_final_keyword_to_query"] = add_final_keyword_to_query
        return self._generic_getter('beacon_api_eth_v1_events_attestation', **kwargs)
 
    def get_proposer(self, **kwargs) -> Any:
        return self._generic_getter('canonical_beacon_proposer_duty', **kwargs)
    
    def get_reorgs(self, **kwargs) -> Any:
        if not "columns" in kwargs:
            kwargs["columns"] = "(slot-depth) as reorged_slot"
        potential_reorgs = self.data_retriever.get_data(
            data_table='beacon_api_eth_v1_events_chain_reorg',
            **kwargs
        )
        if "reorged_slot" not in potential_reorgs.columns:
            potential_reorgs.columns = ["reorged_slot"]
            
        if "slot" in kwargs:
            slot = kwargs["slot"]
            kwargs["slot"] = [slot[0]-32, slot[-1]+31] if isinstance(slot, list) else slot
        else:
            kwargs["slot"] = None
        
        kwargs = {i: j for i, j in kwargs.items() if i != "columns"}
        kwargs["columns"] = "slot"
        kwargs["limit"] = None
        missed = self.get_missed_slots( 
            **kwargs
        )
        reorgs = sorted(set(potential_reorgs["reorged_slot"].tolist()).intersection(missed))
        return pd.DataFrame(reorgs, columns=["slot"])
    
    def get_slots(self, add_missed: bool = True, **kwargs) -> Any:
                
        df = self._generic_getter('canonical_beacon_block', **kwargs)
          
        if add_missed:
            missed = self.get_missed_slots(canonical=df)
            missed_df = pd.DataFrame(missed, columns=['slot'])

            _d = 999999999
            for col, dtype in df.dtypes.items():
                if col != 'slot':  
                    if pd.api.types.is_numeric_dtype(dtype):
                        missed_df[col] = _d if "index" in col else 0
                    else:
                        missed_df[col] = "missed"

            df = pd.concat([df, missed_df], ignore_index=True)
            
            # Add proposer_index for slots that were missed
            if "proposer_index" in df.columns:
                _c = "proposer_validator_index"
                _c1 = "proposer_index"
                _p = self.get_proposer(slot=[int(df.slot.min()), int(df.slot.max()+1)], columns=f"slot,{_c}")
                df[_c1] = df.apply(
                    lambda x: _p[_p["slot"] == x["slot"]][_c].values[0] if x[_c1] == _d else x[_c1], 
                    axis=1
                )
        if "orderby" in kwargs and "," not in kwargs["orderby"]:
            df.sort_values(kwargs["orderby"], inplace=True)
        return df 
 
    def get_missed_slots(self, canonical: Optional = None, **kwargs) -> Any:
        slot = kwargs.get("slot")
        if not slot is None:
            kwargs["slot"] = [slot[0], slot[-1]] if isinstance(slot, list) else slot
        
        if not "add_missed" in kwargs:
            kwargs["add_missed"] = False
        if canonical is None:
            canonical = self.get_slots( 
                **kwargs
            )      
        missed = set(range(canonical.slot.min(), canonical.slot.max()+1)) - set(canonical.slot.unique().tolist())
        return missed
    
    def get_duties(
        self, 
        columns: Optional[str] = "*",
        **kwargs
    ) -> Any:
        required_columns = ["slot", "validators"]            
        kwargs["columns"] = self.clean_columns(columns, required_columns)
        committee = self._generic_getter('beacon_api_eth_v1_beacon_committee', **kwargs)
        committee["validators"] = committee["validators"].apply(lambda x: eval(x))
        duties = pd.DataFrame(columns=["slot", "validators"])
        for i in committee.slot.unique():
            _committee = committee[committee["slot"] == i]
            all_validators = sorted([item for sublist in _committee["validators"] for item in sublist])
            temp_df = pd.DataFrame({"slot": [i] * len(all_validators), "validators": all_validators})
            duties = pd.concat([duties, temp_df], ignore_index=True).drop_duplicates()
        return duties.reset_index(drop=True)
    
    def get_checkpoints(self, slot: int):
        epoch_start_slot = int(slot // 32 * 32)
        last_epoch_start_slot = int(epoch_start_slot - 32)
        slots = self.get_slots(
            slot=[last_epoch_start_slot - 32, epoch_start_slot + 32], 
            columns="slot,block_root", 
            orderby="slot",
            add_missed=False
        )
        head, target, source = [None]*3
        
        _slot = slot
        while head == None:
            if len(slots[slots["slot"] == _slot]) > 0:
                head = slots[slots["slot"] == _slot]["block_root"].values[0]
            else:
                _slot -= 1
        
        _slot = epoch_start_slot
        while target == None:
            if len(slots[slots["slot"] == _slot]) > 0:
                target = slots[slots["slot"] == _slot]["block_root"].values[0]
            else:
                _slot -= 1
                
        _slot = last_epoch_start_slot
        while source == None:
            if len(slots[slots["slot"] == _slot]) > 0:
                source = slots[slots["slot"] == _slot]["block_root"].values[0]
            else:
                _slot -= 1
                
        return head, target, source            
    
    def get_elaborated_attestations(
        self, 
        slot: Optional[Union[int, List[int]]] = None, 
        columns: str = "*",
        what: str = "source,target,head", 
        orderby: Optional[str] = "slot", 
        only_status: Optional[str] = "correct,failed,offline",
        **kwargs
    ) -> Any:

        if not isinstance(slot, list):
            slot = [slot, slot + 1]
        
        required_columns = ["slot", "source_root", "target_root", "validators", "beacon_block_root"]
        
        kwargs["slot"] = [slot[0]//32 * 32, slot[-1]//32 * 32 + 32]
        kwargs["columns"] = self.clean_columns(columns, required_columns)
        kwargs["orderby"] = orderby
        attestations = self.get_attestation(**kwargs)

        kwargs["columns"] = "slot, validators"
        kwargs["limit"] = None
        kwargs["orderby"] = None
        duties = self.get_duties(**kwargs) 

        # Initialize empty list to store all status data
        status_data = []

        # Process each slot
        for _slot in tqdm(sorted(attestations.slot.unique()), desc="Processing slots"):
            head, target, source = self.get_checkpoints(_slot)
            _attestations = attestations[attestations["slot"] == _slot]
            _duties = duties[duties["slot"] == _slot]
            assert len(_duties) > 0, "Something wrong with retrieving duties."
            _all = set(_duties.validators.tolist())
            voting_validators = set(_attestations.validators.tolist())

            def process_vote(vote_type: str, root_value: str) -> None:
                correct = set(_attestations.loc[_attestations[f"{vote_type}_root"] == root_value, 'validators'])
                correct = correct.intersection(_all)
                failing_validators = _all.intersection(voting_validators) - correct
                offline_validators = _all - voting_validators
                
                if "correct" in only_status:
                    status_data.extend([(_slot, v, "correct", vote_type) for v in correct])
                if "failed" in only_status:
                    status_data.extend([(_slot, v, "failed", vote_type) for v in failing_validators])
                if "offline" in only_status:
                    status_data.extend([(_slot, v, "offline", vote_type) for v in offline_validators])

            if "source" in what:
                process_vote("source", source)
            if "target" in what:
                process_vote("target", target)
            if "head" in what:
                process_vote("beacon_block", head)

        final_df = pd.DataFrame(status_data, columns=["slot", "validator", "status", "vote_type"]).sort_values("slot")
        final_df = final_df.drop_duplicates().reset_index(drop=True)

        return final_df  
 
    def get_beacon_block_v2(self, **kwargs) -> Any:
        block = self._generic_getter('beacon_api_eth_v2_beacon_block', **kwargs)
        return block

    def get_block_size(self, orderby: Optional[str] = "slot", **kwargs) -> Any:
        if isinstance(slots, int):
            slots = [slots, slots+1]
        if columns == None:
            columns = []
            
        required_columns = ["slot", "block_total_bytes_compressed", "block_total_bytes", "execution_payload_blob_gas_used"]
        kwargs["slot"] = [slots[0], slots[-1]] if isinstance(slots, list) else slots
        kwargs["columns"] = self.clean_columns(columns, required_columns)
        
        sizes = self.get_slots(**kwargs) 
        if "execution_payload_blob_gas_used" in sizes.columns:
            sizes["blobs"] = sizes["execution_payload_blob_gas_used"] // 131072
            sizes.drop("execution_payload_blob_gas_used", axis=1, inplace=True)
        return sizes
    
    def get_blob_events(self, **kwargs) -> Any:
        return self._generic_getter('beacon_api_eth_v1_events_blob_sidecar', **kwargs)
    
    def get_blobs(self, **kwargs) -> Any:
        return self._generic_getter('canonical_beacon_blob_sidecar', **kwargs)
 
    def get_withdrawals(self, **kwargs) -> Any:
        """
        Retrieve withdrawal data for a given slot.

        Parameters:
        -----------
        slot : Optional[int]
            Slot number to retrieve data for.
        columns : Optional[str]
            Columns to retrieve, default is "*".
        where : Optional[str]
            Optional condition to filter data.
        time_interval : Optional[str]
            Time range to filter data.
        network : str
            Network name, default is "mainnet".
        max_retries : int
            Number of retry attempts, default is 1.
        groupby : str
            Grouping condition.
        orderby : Optional[str]
            Column to order by.
        final_condition : Optional[str]
            Final filter condition.
        limit : int
            Limit the number of rows retrieved.
        store_result_in_parquet : bool
            If true, store result in a Parquet file.
        custom_data_dir : str
            Directory to store Parquet data.

        Available columns (documentation only):
        """
        return self._generic_getter('canonical_beacon_block_withdrawal', **kwargs)

    def create_method_table_mapping(self):
        """
        Dynamically creates the method-to-table mapping by inspecting each method
        and extracting the table name passed to self._get_data().
        """
        method_table_mapping = {}

        # Get all the methods in the class
        for method_name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            # Inspect the method's source code
            source = inspect.getsource(method)
            
            # Dedent the source code to remove unnecessary indentation
            dedented_source = textwrap.dedent(source)
            
            # Extract the table name from the source code
            table_name = self.extract_table_name_from_source(dedented_source)
            if table_name:
                method_table_mapping[method_name] = table_name

        return method_table_mapping
    
    def extract_table_name_from_source(
        self, 
        source: str, 
        current_func_name: str = None, 
        depth: int = 0, 
        max_depth: int = 5
    ) -> Optional[str]:
        """
        Extracts the table name from the method's source code by looking for
        the first argument or keyword argument `data_table` in the call to 
        self._get_data() or self.get_data(), with recursion to resolve nested calls.

        Recurses up to `max_depth` times when encountering function calls.
        """
        try:
            # Check if we've exceeded the maximum recursion depth
            if depth > max_depth:
                return None

            # Parse the source code into an abstract syntax tree (AST)
            tree = ast.parse(source)

            # Iterate over all nodes in the AST
            for node in ast.walk(tree):
                # Look for calls to self._get_data() or self.get_data()
                if isinstance(node, ast.Call) and isinstance(node.func, ast.Attribute):

                    if node.func.attr in ['_generic_getter', 'get_data', '_get_data']:
                        # Check if the table name is passed as a keyword argument
                        for keyword in node.keywords:
                            if keyword.arg == 'data_table':
                                if isinstance(keyword.value, ast.Constant):
                                    table_name = keyword.value.value
                                    return table_name
                            
                        # Handle positional arguments
                        if node.args and len(node.args) > 0:
                            first_arg = node.args[0]

                            if isinstance(first_arg, ast.Constant):
                                table_name = first_arg.value
                                return table_name
                          
                    # If it's not a get_data or _get_data call, check which function is being called
                    elif isinstance(node.func.value, ast.Name):
                        called_function = node.func.attr

                        # Load the source of the called function and recurse
                        try:
                            called_func_source = inspect.getsource(getattr(self, called_function))
                            # Clean up the source to handle any potential indentation issues
                            cleaned_source = textwrap.dedent(called_func_source).strip()
                            return self.extract_table_name_from_source(
                                cleaned_source, called_function, 
                                depth + 1, 
                                max_depth
                            )
                        except (AttributeError, OSError) as e:
                            pass
               
        except Exception as e:
            pass
        
        return None


    def update_all_column_docs(self):
        """
        Updates the docstrings of all high-level methods by using the table information stored in DocsScraper.
        """
        # Fetch the table info from DocsScraper (already stored during initialization)
        all_table_info = {table: self.get_docs(table, False) for table in self.method_table_mapping.values()}

        # Iterate through methods and update their docstrings
        for method_name, table_name in self.method_table_mapping.items():
            method = getattr(self, method_name, None)
            table_info = all_table_info.get(table_name)

            if table_info is not None and not table_info.empty:
                columns_doc = "\n".join([f"  - {col}" for col in table_info['Column']])
            else:
                columns_doc = "  No columns available."

            # Use a helper function to avoid closure issues
            self._wrap_method_with_columns(method_name, method, table_name, columns_doc)

    def _wrap_method_with_columns(
        self, 
        method_name: str, 
        method: Optional[Callable], 
        table_name: str, 
        columns_doc: str
    ) -> None:
        """
        Helper function to wrap a method with a new docstring that includes the available columns.

        :param method_name: The name of the method to wrap.
        :param method: The method to be wrapped.
        :param table_name: The name of the table associated with the method.
        :param columns_doc: The documentation string containing the available columns.
        """
        # Define a wrapper function that maintains the original method behavior
        def method_wrapper(*args, **kwargs):
            return method(*args, **kwargs)

        # Update the docstring of the wrapper function
        method_wrapper.__doc__ = (method.__doc__ or "") + f"\nAvailable columns for {table_name}:\n{columns_doc}\n"

        # Replace the original method with the wrapped version
        setattr(self, method_name, method_wrapper)    

    def verify_columns(self, columns: str = None, table: str = None):
        if columns is None or table is None:
            return True
        assert type(columns) == str
        if columns == "*":
            return True
        if "," not in columns:
            columns += ","
        
        existing_columns = self.get_docs(table, False)
        if existing_columns is None:
            return True
        if "Column" in existing_columns.columns:
            existing_columns = existing_columns["Column"].tolist()
        else:
            return True
        for c in [i for i in columns.split(",") if i != ""]:
            _c = self.helpers.extract_inside_brackets(c.strip())
            if _c not in existing_columns:
                if _c == "" or _c == " ":
                    continue
                print("\n" + f"{_c.strip()} not in {table} with columns:" + '\n'.join(existing_columns))
                print("\nExisting columns: " + '\n'.join(self.get_docs(table, False)['Column'].to_list()))
                return False
        return True
    
    def clean_columns(self, columns: str, required_columns: List[str]) -> str:
        columns_list = [i.strip() for i in columns.split(",") if i != "*"]
        return ",".join(list(dict.fromkeys(columns_list + required_columns)))

    def preview_result(self, func: Callable[..., Any], limit: int = 100, **kwargs) -> Any:
        kwargs['limit'] = limit
        return func(**kwargs)
    
    def __str__(self) -> str:
        return f"PyXatu(config_path={self.config_path}, user={self.clickhouse_user}, url={self.clickhouse_url})"

    def __repr__(self) -> str:
        return self.__str__()


import time
import logging
from functools import wraps
from typing import Any, Callable, TypeVar

TABLES = {
    "beacon_api_eth_v1_events_block": "default.beacon_api_eth_v1_events_block",
    "canonical_beacon_proposer_duty": "default.canonical_beacon_proposer_duty",
    "beacon_api_eth_v1_events_chain_reorg": "default.beacon_api_eth_v1_events_chain_reorg",
    "canonical_beacon_block": "default.canonical_beacon_block",
    "canonical_beacon_elaborated_attestation": "default.canonical_beacon_elaborated_attestation",
    "beacon_api_eth_v1_events_attestation": "default.beacon_api_eth_v1_events_attestation",
    "beacon_api_eth_v1_events_blob_sidecar": "default.beacon_api_eth_v1_events_blob_sidecar",
    "canonical_beacon_blob_sidecar": "default.canonical_beacon_blob_sidecar",
    "beacon_api_eth_v1_beacon_committee": "default.beacon_api_eth_v1_beacon_committee",
    "canonical_beacon_block_withdrawal": "default.canonical_beacon_block_withdrawal",    
    "beacon_api_eth_v2_beacon_block": "default.beacon_api_eth_v2_beacon_block",    
    
}

GENESIS_TIME_ETH_POS = 1606824023
SECONDS_PER_SLOT = 12

CONSTANTS = {
    "TABLES": TABLES,
    "GENESIS_TIME_ETH_POS": GENESIS_TIME_ETH_POS,
    "SECONDS_PER_SLOT": SECONDS_PER_SLOT
}

F = TypeVar('F', bound=Callable[..., Any])

def retry_on_failure(max_retries: int = 1, initial_wait: float = 1.0, backoff_factor: float = 2.0) -> Callable[[F], F]:
    """Decorator to retry a function if an exception occurs."""
    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            attempt = 0
            wait_time = initial_wait
            while attempt < max_retries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    logging.warning(f"Attempt {attempt + 1}/{max_retries} failed: {e}. Retrying in {wait_time} seconds...")
                    time.sleep(wait_time)
                    wait_time *= backoff_factor
                    attempt += 1
            logging.error(f"Max retries reached. Failed to complete operation.")
            return None
        return wrapper  # type: ignore
    return decorator


import unittest
from unittest.mock import MagicMock, patch
import time
from pyxatu.utils import retry_on_failure

class TestRetryOnFailure(unittest.TestCase):
    
    @patch('pyxatu.utils.logging')
    def test_successful_execution(self, mock_logging):
        @retry_on_failure()
        def mock_function():
            return "Success"
        
        self.assertEqual(mock_function(), "Success")
    
    @patch('pyxatu.utils.logging')
    def test_retry_on_failure(self, mock_logging):
        mock_func = MagicMock(side_effect=[Exception("Fail"), "Success"])
        
        @retry_on_failure(max_retries=2)
        def mock_function():
            return mock_func()
        
        self.assertEqual(mock_function(), "Success")
        self.assertEqual(mock_func.call_count, 2)
    
    @patch('time.sleep', return_value=None)
    @patch('pyxatu.utils.logging')
    def test_retry_max_attempts(self, mock_logging, mock_sleep):
        mock_func = MagicMock(side_effect=Exception("Fail"))
        
        @retry_on_failure(max_retries=3)
        def mock_function():
            return mock_func()
        
        self.assertIsNone(mock_function())
        self.assertEqual(mock_func.call_count, 3)

if __name__ == '__main__':
    unittest.main()


import os
import pandas as pd
from typing import Optional, List, Any
import time

class DataRetriever:
    def __init__(self, client, tables):
        self.client = client
        self.tables = tables

    def get_data(self, data_table: str, slot: Optional[int] = None, columns: str = "*", 
                 where: Optional[str] = None, time_interval: Optional[str] = None, network: str = "mainnet", 
                 groupby: str = None, orderby: Optional[str] = None, final_condition: Optional[str] = None, 
                 limit: int = None, store_result_in_parquet: bool = None, custom_data_dir: str = None,
                 add_final_keyword_to_query: bool = True) -> Any:
        if columns != "*" and isinstance(columns, list):
            columns = ",".join(columns)
        table = self.tables.get(data_table)
        if not table:
            raise ValueError(
                f"Data table '{data_table}' is not valid. Please use one of the following: {', '.join(self.tables.keys())}"
            )
        result = self.client.fetch_data(
            table=table, 
            slot=slot, 
            columns=columns, 
            where=where, 
            time_interval=time_interval, 
            network=network, 
            groupby=groupby, 
            orderby=orderby, 
            final_condition=final_condition,
            limit=limit,
            add_final_keyword_to_query=add_final_keyword_to_query
        )
        if not result is None:
            if store_result_in_parquet:
                self.store_result_to_disk(result, custom_data_dir)
            else:
                return result

        
        
    def store_result_to_disk(self, result, custom_data_dir: str = None):
        if custom_data_dir is None:
            custom_data_dir = './output_data/output.parquet'
           
        directories = custom_data_dir.split("/")[:-1]
        for i in range(len(directories)):
            directory = directories[i]
            if directory == ".":
                continue
            if not os.path.isdir(directory):
                os.mkdir(directory)
            print(f"Created directory `{directory}`")

        if os.path.exists(custom_data_dir):
            timestamp = int(time.time())
            new_file_path = f"{os.path.splitext(custom_data_dir)[0]}_{timestamp}{os.path.splitext(custom_data_dir)[1]}"
            
            print(f"File '{custom_data_dir}' already exists. Using a new location: '{new_file_path}'")
            custom_data_dir = new_file_path  # Automatically suggest the new filename
        
        # Save the result to the specified or new file
        result.to_parquet(custom_data_dir, index=True)
        print(f"File saved at '{custom_data_dir}'")



import unittest
from unittest.mock import patch, MagicMock
from pyxatu.core import PyXatu
import os
import pandas as pd
from pathlib import Path


class TestPyXatu(unittest.TestCase):

    def setUp(self):
        # Patch the ClickhouseClient and DataRetriever
        self.patcher_client = patch('pyxatu.core.ClickhouseClient', autospec=True)
        self.patcher_retriever = patch('pyxatu.core.DataRetriever', autospec=True)

        # Start the patchers
        self.mock_client = self.patcher_client.start()
        self.mock_retriever = self.patcher_retriever.start()

        # Mock instances
        self.mock_client_instance = self.mock_client.return_value
        self.mock_retriever_instance = self.mock_retriever.return_value

        # Instantiate PyXatu with mocks
        self.pyxatu = PyXatu(config_path=None, use_env_variables=True)

    def tearDown(self):
        self.patcher_client.stop()
        self.patcher_retriever.stop()
    
    @patch.dict(os.environ, {
        "CLICKHOUSE_USER": "test_user",
        "CLICKHOUSE_PASSWORD": "test_password",
        "CLICKHOUSE_URL": "http://test-url"
    })
    def test_init_with_env_variables(self):
        # Reset mock before asserting to remove previous calls
        self.mock_client.reset_mock()

        # Initialize PyXatu with environment variables
        pyxatu_instance = PyXatu(config_path=None, use_env_variables=True)

        # Now assert the ClickhouseClient was called with the correct parameters
        self.mock_client.assert_called_once_with(
            "http://test-url", "test_user", "test_password"
        )

        # Ensure that DataRetriever is initialized
        self.assertIsNotNone(pyxatu_instance.data_retriever)

    @patch('pyxatu.core.PyXatu.read_clickhouse_config_locally', return_value=("http://test-url", "user", "pass"))
    def test_init_with_config_file(self, mock_read_config):
        
        default_path = os.path.join(Path.home(), '.pyxatu_config.json')
        
        # Reset the mock before asserting
        self.mock_client.reset_mock()

        # Initialize PyXatu with config file
        pyxatu_instance = PyXatu(config_path=default_path, use_env_variables=False)

        # Ensure the config file method was called
        mock_read_config.assert_called_once()

        # Now assert the ClickhouseClient was called with the config file values
        self.mock_client.assert_called_once_with(
            "http://test-url", "user", "pass"
        )

    def test_get_blockevent(self):
        # Set up the return value of get_data from DataRetriever
        self.mock_retriever_instance.get_data.return_value = 'mock_result'

        result = self.pyxatu.get_blockevent(slot=12345)

        # Assert that DataRetriever.get_data was called with correct arguments
        self.mock_retriever_instance.get_data.assert_called_once_with(
            'beacon_api_eth_v1_events_block',
            slot=12345,
        )
        self.assertEqual(result, 'mock_result')

    def test_get_reorgs(self):
        # Mock DataRetriever.get_data for reorgs and canonical slots
        self.mock_retriever_instance.get_data.side_effect = [
            pd.DataFrame({'reorged_slot': [9000000, 9000001]}),  # Mock reorgs data
            pd.DataFrame({'slot': [8999999, 9000000, 9000001, 9000002, 9000003]})  
        ]

        # Call the method under test
        result = self.pyxatu.get_reorgs(slot=[9000000, 9000001])

        # Ensure that get_data was called twice: once for reorgs and once for canonical slots
        self.assertEqual(self.mock_retriever_instance.get_data.call_count, 2)

        # Verify that the result contains the correct reorg slots
        self.assertEqual(result.to_string(), pd.DataFrame([], columns=["slot"]).to_string())


    def test_get_reorgs_no_reorgs(self):
        # Mock DataRetriever.get_data for reorgs (no reorgs in the data)
        self.mock_retriever_instance.get_data.side_effect = [
            pd.DataFrame({'reorged_slot': []}),  # No reorgs
            pd.DataFrame({'slot': [8999999, 9000000, 9000001, 9000002, 9000003]})  # Mock canonical slots data
        ]

        # Call the method under test with slots that do not have reorgs
        result = self.pyxatu.get_reorgs(slot=[9000000, 9000001])

        # Ensure that the result is an empty list, as there are no reorgs
        self.assertEqual(result.to_string(), pd.DataFrame([], columns=["slot"]).to_string())

    def test_get_reorgs_missing_canonical_slots(self):
        # Mock DataRetriever.get_data for reorgs and canonical slots
        # Simulate missing canonical slots
        self.mock_retriever_instance.get_data.side_effect = [
            pd.DataFrame({'reorged_slot': [9000000, 9000001]}),  # Mock reorgs data
            pd.DataFrame({'slot': [8999999, 9000002, 9000003]})  # Missing canonical slots 9000000 and 9000001
        ]

        # Call the method under test with slots
        result = self.pyxatu.get_reorgs(slot=[9000000, 9000001])

        # Ensure that the correct reorg slots are identified
        self.assertEqual(result.to_string(), pd.DataFrame([9000000, 9000001], columns=["slot"]).to_string())

    def test_execute_query(self):
        # Set up the return value of execute_query from ClickhouseClient
        self.mock_client_instance.execute_query.return_value = 'mock_query_result'

        result = self.pyxatu.execute_query("SELECT * FROM test_table")

        # Assert that ClickhouseClient.execute_query was called with correct arguments
        self.mock_client_instance.execute_query.assert_called_once_with(
            "SELECT * FROM test_table", "*"
        )
        self.assertEqual(result, 'mock_query_result')

if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch, MagicMock
import pandas as pd
import os
import json
from pyxatu.client import ClickhouseClient


CONFIG_PATH = os.path.expanduser("~/.pyxatu_config.json")

print(CONFIG_PATH)
with open(CONFIG_PATH, 'r') as file:
    config = json.load(file)
    clickhouse_user = config.get("CLICKHOUSE_USER", "default_user")
    clickhouse_password = config.get("CLICKHOUSE_PASSWORD", "default_password")
    url = config.get("CLICKHOUSE_URL", "http://localhost")
        
os.environ["CLICKHOUSE_URL"] = url
os.environ["CLICKHOUSE_USER"] = clickhouse_user
os.environ["CLICKHOUSE_PASSWORD"] = clickhouse_password


class TestClickhouseClient(unittest.TestCase):
    
    @patch.dict(os.environ)
    def setUp(self):
        # Initialize the ClickhouseClient with the environment variables
        self.client = ClickhouseClient(
            url=os.getenv("CLICKHOUSE_URL"),
            user=os.getenv("CLICKHOUSE_USER"),
            password=os.getenv("CLICKHOUSE_PASSWORD")
        )

    @patch('requests.get')
    def test_execute_query_success(self, mock_get):
        mock_response = MagicMock()
        mock_response.raise_for_status.return_value = None
        mock_response.text = "value1\tvalue2"
        mock_get.return_value = mock_response
        
        result = self.client.execute_query("SELECT * FROM test_table")
        
        expected_df = pd.DataFrame([["value1", "value2"]], columns=[0, 1])
        pd.testing.assert_frame_equal(result, expected_df)

    @patch('requests.get')
    @patch('pyxatu.utils.logging')
    def test_execute_query_failure(self, mock_logging, mock_get):
        mock_get.side_effect = Exception("Request Failed")   
        result = self.client.execute_query("SELECT * FROM test_table")   
        self.assertIsNone(result)

    @patch('pyxatu.client.ClickhouseClient._build_query')
    @patch('pyxatu.client.ClickhouseClient.execute_query')
    def test_fetch_data(self, mock_execute_query, mock_build_query):
        mock_build_query.return_value = 'SELECT column1 FROM some_table'
        mock_execute_query.return_value = pd.DataFrame({'column1': [1, 2, 3]})

        # Call the real fetch_data method and test the result
        result = self.client.fetch_data(
            table='test_table', 
            slot=1234, 
            columns=['col1', 'col2'], 
            where=None, 
            time_interval=None, 
            network='mainnet', 
            orderby=None, 
            final_condition=None, 
            limit=None
        )

        # Assert that mock_build_query and mock_execute_query were called
        mock_build_query.assert_called_once()
        mock_execute_query.assert_called_once_with('SELECT column1 FROM some_table', ['col1', 'col2'])

        # Verify the result is as expected
        self.assertIsInstance(result, pd.DataFrame)
        self.assertFalse(result.empty)

    @patch('requests.get')
    def test_execute_query_with_slot(self, mock_get):
        # Mock the response
        mock_response = MagicMock()
        mock_response.raise_for_status.return_value = None
        mock_response.text = "9700000\t2024-08-09 17:20:23"
        mock_get.return_value = mock_response

        self.setUp()

        expected_query = "SELECT DISTINCT slot, slot_start_date_time FROM default.canonical_beacon_block WHERE slot = 9700000 AND slot_start_date_time = '2024-08-09 17:20:23'"

        result = self.client.execute_query(expected_query)
        mock_get.assert_called_once_with(
            self.client.url,
            params={'query': expected_query},
            auth=self.client.auth,
            timeout=1500
        )

        # Verify the result as expected
        expected_df = pd.DataFrame([[9700000, '2024-08-09 17:20:23']], columns=["slot", "slot_start_date_time"])
        
        pd.testing.assert_frame_equal(result, expected_df)


if __name__ == '__main__':
    unittest.main()



import os
import logging
import pandas as pd

class ValidatorGadget:
    
    def __init__(self):
        self.mapping = self.load_validator_mapping()
    
    def _download_validator_mapping(self):
        df = pd.read_parquet("https://storage.googleapis.com/public_eth_data/openethdata/validator_data.parquet.gzip")
        df["validator_id"] = df["validator_id"].astype(int)
        df["lido_node_operator"] = df["lido_node_operator"].apply(lambda x: x.lower())
        df["label"] = df["label"].apply(lambda x: x.lower())
        df.to_parquet("validator_mapping.parquet", index=False)
        logging.info("Validator mapping downloaded and stored to `./validator_mapping.parquet`")
    
    def _build_validator_mapping(self):
        
        validators = self.load_validators_locally()
        
        self.add_coinbase_validators()
        
        self.add_kiln_validators()
        
        self.add_binance_validators()
        
        self.add_rocketpool_validators()
        
        
    def _add_coinbase_validators(self):
        pass

    def _add_kiln_validators(self):
        pass

    def _add_binance_validators(self):
        pass

    def _add_rocketpool_validators(self):
        pass
        
        
    def load_validator_mapping(self):
        if not os.path.isfile("validator_mapping.parquet"):
            logging.warning("No validator mapping found locally")
            self._download_validator_mapping()
            return self.load_validator_mapping()
        return pd.read_parquet("validator_mapping.parquet")
         


import re
import time
from datetime import datetime, timezone
from typing import List

from pyxatu.utils import CONSTANTS


class PyXatuHelpers:
    
    def get_slot_datetime(self, slot: int) -> int:
        slot_timestamp = CONSTANTS["GENESIS_TIME_ETH_POS"] + (slot * CONSTANTS["SECONDS_PER_SLOT"])
        slot_datetime = datetime.fromtimestamp(slot_timestamp, tz=timezone.utc)
        return slot_datetime.strftime('%Y-%m-%d %H:%M:%S')     
    
    def get_slot_timestamp(self, slot: int) -> int:
        slot_timestamp = CONSTANTS["GENESIS_TIME_ETH_POS"] + (slot * CONSTANTS["SECONDS_PER_SLOT"])
        slot_datetime = datetime.fromtimestamp(slot_timestamp, tz=timezone.utc)
        return int(slot_datetime.timestamp())     

    def get_time_in_slot(self, slot: int, ts: int = None) -> int:
        return (ts - self.get_slot_timestamp(slot = slot)*1000)/1000
    
    def date_string_to_timestamp(self, date_string: str):
        dt_obj = datetime.strptime(date_str, '%Y-%m-%d %H:%M:%S.%f')
        utc_timestamp = dt_obj.timestamp()
        returnutc_timestamp
        
    def slot_to_time(self, slot: int, _format="%Y-%m-%d %H:%M:%S"):
        """Convert a slot number to a formatted time string."""
        timestamp = 1606824023 + slot * 12
        dt_object = datetime.utcfromtimestamp(timestamp)
        return dt_object.strftime(_format)
    
    def slot_to_day(self, slot: int):
        return self.slot_to_time(slot, "%Y-%m-%d")
    
    def get_current_ethereum_slot(self):
        current_timestamp = int(time.time()) - 64*12
        slot_number_at_known_timestamp = 8000000
        known_timestamp = 1702824023  # Timestamp for slot 8000000 in UTC
        seconds_per_slot = 12
        current_slot = slot_number_at_known_timestamp + (current_timestamp - known_timestamp) // seconds_per_slot
        return current_slot
    
    def extract_inside_brackets(self, input_string: str = None):
        match = re.search(r'\((.*?)\)', input_string)
        if match:
            return match.group(1)
        else:
            return input_string

    def check_types(self, values_list: List[str], types_list: List[type]):
        """
        Checks whether the types of elements in values_list match the expected types in types_list.

        :param values_list: List of values to check.
        :param types_list: List of types corresponding to values in values_list.
        :return: True if all values match their respective types, False otherwise.
        :raises ValueError: If the lengths of values_list and types_list do not match.
        """
        if len(values_list) != len(types_list):
            raise ValueError("The length of values_list and types_list must be the same.")

        types_list = [str(i) for i in types_list]
        
        for value, expected_type in zip(values_list, types_list):
            if str(type(value)) not in expected_type:
                print(f"{type(value)} != {expected_type}")
                return False
        return True


import requests
from bs4 import BeautifulSoup
import pandas as pd

class DocsScraper:
    
    def __init__(self):
        # Dictionary to store table information for all tables
        self.tables = dict()
        self.scrape_all_tables()
        
    def _get_table_info(self, soup, table_header):
        """
        Extracts table information given the table header element.
        """
        table = table_header.find_next('table')
        if not table:
            return None
        
        columns = []
        types = []
        descriptions = []

        rows = table.find_all('tr')

        for row in rows[1:]:  # Skip the header row
            cells = row.find_all('td')
            if len(cells) == 3:
                columns.append(cells[0].get_text(strip=True))
                types.append(cells[1].get_text(strip=True))
                descriptions.append(cells[2].get_text(strip=True))

        # Creating a pandas DataFrame
        data = pd.DataFrame({
            'Column': columns,
            'Type': types,
            'Description': descriptions
        })

        return data
    
    def scrape_all_tables(self):
        """
        Scrapes all tables from the given URL and stores their dataframes in a dictionary.
        """
        url = "https://ethpandaops.io/data/xatu/schema/"
        response = requests.get(url)
        soup = BeautifulSoup(response.content, 'html.parser')

        # Find all sections containing tables
        headers = soup.find_all('h2', class_='relative group')
        for header in headers:
            table_name = header.get_text(strip=True).replace("#", "").split()[0]
            if table_name.lower() != "tables":  # Skip the general 'Tables' header
                table_data = self._get_table_info(soup, header)
                if table_data is not None:
                    self.tables[table_name] = table_data

    def get_table_info(self, table_name: str = None):
        """
        Returns the table info as a pandas DataFrame for the specified table.
        If the table_name is None, it returns a list of all available tables.
        """
        if table_name is None:
            return list(self.tables.keys())  # Return all available table names

        if table_name in self.tables:
            return self.tables[table_name]
        
        print(f"Table {table_name} not found.")
        return None


import time
import logging
import requests
import pandas as pd
from io import StringIO
from datetime import datetime, timezone
from typing import Optional, List, Any
from requests.auth import HTTPBasicAuth

from pyxatu.utils import retry_on_failure, CONSTANTS
from pyxatu.helpers import PyXatuHelpers

class ClickhouseClient:
    def __init__(self, url: str, user: str, password: str, timeout: int = 1500, helper: Any = None) -> None:
        self.url = url
        self.auth = HTTPBasicAuth(user, password)
        self.timeout = timeout
        self.helpers = helper or PyXatuHelpers()

    @retry_on_failure()
    def execute_query(self, query: str, columns: Optional[str] = "*") -> pd.DataFrame:
        logging.info(f"Executing query: {query}")
        start_time = time.time()
        response = requests.get(
            self.url,
            params={'query': query},
            auth=self.auth,
            timeout=self.timeout
        )
        logging.info(f"Query executed in {time.time() - start_time:.2f} seconds")
        response.raise_for_status()
        if "DISTINCT" in query.upper():
            potential_columns = query.split("FROM")[0].split("DISTINCT")[1].strip()
        else:
            potential_columns = query.split("FROM")[0].split("SELECT")[1].strip()
        if potential_columns != "*" and "," in potential_columns:
            potential_columns = ",".join([i.split("as ")[-1].strip() if "as " in i else i.strip() for i in potential_columns.split(",")])
        elif potential_columns != "*":
            potential_columns = [i.split("as ")[-1].strip() if "as " in i else i.strip() for i in [potential_columns]] 
            
        if response.text == "":
            logging.info("No data for query")
            return None
            
        return self._parse_response(response.text, columns, potential_columns)

    def _parse_response(self, response_text: str, columns: Optional[str] = "*", potential_columns: Optional[str] = None) -> pd.DataFrame:
        """Converts response text to a Pandas DataFrame and assigns column names if provided."""
        df = pd.read_csv(StringIO(response_text), sep='\t', header=None)
        if columns and columns != "*":
            columns = [col.strip() for col in columns.split(',')]
            df.columns = [i.split("as ")[-1] if " as " in i else i for i in columns]

        elif potential_columns and potential_columns != "*":
            df.columns = [col.strip() for col in potential_columns.split(",")] 
            
        return df

    def fetch_data(self, table: str, slot: Optional[int] = None, columns: str = '*', where: Optional[str] = None,
                   time_interval: Optional[str] = None, network: str = "mainnet", groupby: Optional[str] = None,
                   orderby: Optional[str] = None, final_condition: Optional[str] = None, limit: int = None,
                   add_final_keyword_to_query: bool = True) -> pd.DataFrame:
        query = self._build_query(
            table = table, 
            slot = slot, 
            columns = columns, 
            where = where, 
            time_interval = time_interval, 
            network = network, 
            groupby = groupby, 
            orderby = orderby, 
            final_condition = final_condition,
            limit = limit,
            add_final_keyword_to_query=add_final_keyword_to_query
        )
        return self.execute_query(query, columns)

    def _build_query(self, table: str, slot: Optional[int], columns: str, where: Optional[str], 
                     time_interval: Optional[str], network: str,  groupby: Optional[str], orderby: Optional[str], 
                     final_condition: Optional[str], limit: int = None, add_final_keyword_to_query: bool = True) -> str:
        query = f"SELECT DISTINCT {columns} FROM {table}"
        if add_final_keyword_to_query:
            query += " FINAL"
        conditions = []
        
        if isinstance(slot, int):
            date_filter = self._get_sql_date_filter(slot=slot)
            conditions.append(date_filter)
            conditions.append(f"slot = {int(slot)}")
        elif slot and isinstance(slot, list) and len(slot) == 2:
            date_filter = self._get_sql_date_filter(slot=slot)
            conditions.append(date_filter)
            conditions.append(f"slot >= {int(slot[0])} AND slot < {int(slot[1])}")
        
        if where: conditions.append(where)
        if time_interval: conditions.append(f"slot_start_date_time > NOW() - INTERVAL '{time_interval}'")
        if network: conditions.append(f"meta_network_name = '{network}'")
        if final_condition: conditions.append(final_condition)
        
        query += f" WHERE {' AND '.join(filter(None, conditions))}"
        
        if groupby: query += f" GROUP BY {groupby}"
        if orderby: query += f" ORDER BY {orderby}"
            
        if limit: query += f" LIMIT {limit}"
        
        return query
    
    def _get_sql_date_filter(self, slot: Optional[int] = None) -> str:
        """
        Returns a SQL-compatible date filter for a given Ethereum PoS slot or a range of slots.
        This is useful to minimize the amount of requested resources on the Xatu backend.

        Args:
            slot (Optional[int]): A single slot number to calculate the date for.
            slots (Optional[List[int]]): A list of two slot numbers to create a range filter.

        Returns:
            str: A SQL date filter in the format 'YYYY-MM-DD HH:MM:SS' or a range of timestamps.

        Raises:
            ValueError: If neither a valid `slot` nor a valid list of `slots` is provided.
        """
        # Case 1: Single slot provided (it must be an integer)
        if isinstance(slot, int):
            slot_date_str = self.helpers.get_slot_datetime(slot)
            slot_date_str_n = self.helpers.get_slot_datetime(slot+1)
            
            return f"slot_start_date_time >= '{slot_date_str}' AND slot_start_date_time < '{slot_date_str_n}'"

        # Case 2: List of two slots provided (must be a list of exactly two integers)
        elif isinstance(slot, list) and len(slot) == 2 and all(isinstance(s, int) for s in slot):
            lower_slot_date_str = self.helpers.get_slot_datetime(slot[0])
            upper_slot_date_str = self.helpers.get_slot_datetime(slot[1])
            return f"slot_start_date_time >= '{lower_slot_date_str}' AND slot_start_date_time < '{upper_slot_date_str}'"

        else:
            print(slot)
            raise ValueError(f"Invalid input: either a valid integer slot or a list of exactly two slots must be provided. Provided input type: {type(slot)}, Slot variable contains: {str(slot)}")



import unittest
from unittest.mock import MagicMock, patch
import pandas as pd
import os
from pyxatu.retriever import DataRetriever

class TestDataRetriever(unittest.TestCase):

    def setUp(self):
        self.client = MagicMock()  # Mock the client
        self.tables = {
            "valid_data_table": "test_table"
        }
        self.retriever = DataRetriever(client=self.client, tables=self.tables)

    def test_get_data_success(self):
        # Mock fetch_data to return a valid DataFrame
        mock_result = pd.DataFrame({'col1': [1, 2], 'col2': ['A', 'B']})
        self.client.fetch_data.return_value = mock_result

        result = self.retriever.get_data(
            data_table="valid_data_table",  # Ensure this matches the keys in `tables`
            slot=1234,
            columns="col1,col2"
        )

        # Use keyword arguments to match the actual call in `get_data`
        self.client.fetch_data.assert_called_once_with(
            table="test_table", 
            slot=1234, 
            columns="col1,col2", 
            where=None, 
            time_interval=None, 
            network="mainnet", 
            groupby=None,
            orderby=None, 
            final_condition=None, 
            limit=None,
            add_final_keyword_to_query=True
        )

        # Ensure the result is as expected
        pd.testing.assert_frame_equal(result, mock_result)

    def test_get_data_invalid_data_table(self):
        # Test for invalid data table
        with self.assertRaises(ValueError) as context:
            self.retriever.get_data(data_table="invalid_data_table")

        self.assertIn("Data table 'invalid_data_table' is not valid", str(context.exception))
        
    @patch('pyxatu.retriever.os.path.exists', return_value=True)
    @patch('pyxatu.retriever.os.path.isdir', return_value=False)
    @patch('pyxatu.retriever.os.mkdir')
    @patch('pyxatu.retriever.time.time', return_value=1234567890)
    @patch('pyxatu.retriever.pd.DataFrame.to_parquet')
    def test_store_result_to_disk(self, mock_to_parquet, mock_time, mock_mkdir, mock_isdir, mock_exists):
        # Mock DataFrame
        mock_result = pd.DataFrame({'col1': [1, 2], 'col2': ['A', 'B']})

        # Test storing data when directory doesn't exist
        custom_dir = './test_output/output.parquet'
        
        self.retriever.store_result_to_disk(mock_result, custom_dir)
        
        # Assert mkdir was called to create directory
        mock_mkdir.assert_called_once_with('test_output')
        
        # Assert result is saved to the file
        expected_file_path = './test_output/output_1234567890.parquet'
        mock_to_parquet.assert_called_once_with(expected_file_path, index=True)

    @patch('pyxatu.retriever.os.path.exists',  side_effect=[True, False])  # Mock os.path.exists
    @patch('pyxatu.retriever.os.path.isdir', return_value=False)  # Mock os.path.isdir to return False
    @patch('pyxatu.retriever.os.mkdir')  # Mock os.mkdir
    @patch('pyxatu.retriever.pd.DataFrame.to_parquet')  # Mock to_parquet
    @patch('pyxatu.retriever.time.time', return_value=1234567890)  # Mock time to return a fixed timestamp
    def test_store_result_to_disk_without_existing_file(self, mock_time, mock_to_parquet, mock_mkdir, mock_isdir, mock_exists):
        # Mock DataFrame
        mock_result = pd.DataFrame({'col1': [1, 2], 'col2': ['A', 'B']})

        # Test storing data when the file doesn't exist
        custom_dir = 'test_output/output.parquet'

        self.retriever.store_result_to_disk(mock_result, custom_dir)

        # Assert mkdir was called to create the directory
        mock_mkdir.assert_called_once_with('test_output')  # Ensure mkdir was called with the correct directory

        # Assert the result is saved to the new file with a timestamp
        expected_file_path = 'test_output/output_1234567890.parquet'
        mock_to_parquet.assert_called_once_with(expected_file_path, index=True)
        
if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

setup(
    name='pyxatu',
    version='1.5',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'pyxatu': ['config.json'],
    },
    install_requires=[
        'requests',
        'pandas',
        'tqdm',
        'bs4',
        'termcolor',
        'fastparquet',
        'click'
    ],
    entry_points={
        'console_scripts': [
            'xatu=pyxatu.cli:cli',
        ],
    },
    author='Toni Wahrstätter',
    author_email='toni@ethereum.org',
    description='A Python interface for the Xatu API',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/nerolation/pyxatu',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
)



import click
import importlib.resources as resources
from pathlib import Path
import shutil
from .core import PyXatu

@click.group()
def cli():
    """Main command group for PyXatu CLI."""
    pass

@click.command()
def setup():
    """CLI command to set up the default configuration in the user's home directory."""
    home = Path.home()
    user_config_path = home / '.pyxatu_config.json'

    try:
        print(f"Looking for config.json in the installed package...")
        default_config_file = resources.files('pyxatu') / 'config.json'
        print(f"Config file found at: {default_config_file}")

        if not user_config_path.exists():
            shutil.copy(default_config_file, user_config_path)
            print(f"Default configuration copied to {user_config_path}.")
            print("Enter your credentials to use PyXatu.")
        else:
            print(f"User configuration already exists at {user_config_path}.")
    except Exception as e:
        print(f"Error during configuration setup: {e}")

@click.command()
@click.argument('query')
@click.option('--config', type=click.Path(exists=True), help='Path to the configuration file', default=None)
@click.option('--columns', multiple=True, help='List of column names')
def query(query, config, columns):
    """Run queries against the Xatu API."""
    xatu = PyXatu(config_path=config)
    result = xatu.request_query(query, columns=columns)
    
    if result is not None:
        print(result)
    else:
        print("Query failed")

# Add both commands to the main CLI group
cli.add_command(setup)
cli.add_command(query)

if __name__ == "__main__":
    cli()
